"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var R = require("ramda");
var definitions_interface_1 = require("../../../definitions.interface");
var util_1 = require("../../../util");
/**
 * @stable [26.07.2019]
 * @param {Partial<IMultiEntity>} initial
 * @returns {IMultiEntity}
 */
exports.multiEntityFactory = function (initial) { return ({
    add: initial.add || [],
    edit: initial.edit || [],
    remove: initial.remove || [],
    source: initial.source || [],
}); };
/**
 * @stable [27.02.2019]
 * @param {IEntity[]} currentEntities
 * @param {number} itemsLimit
 * @returns {TItem[]}
 */
exports.asViewedMultiItemEntities = function (currentEntities, itemsLimit) {
    var result = util_1.generateArray(itemsLimit);
    var actualRelations = exports.toActualMultiItemEntities(currentEntities);
    if (Array.isArray(actualRelations)) {
        var cursor_1 = 0;
        result.forEach(function (_, index) {
            var entity = actualRelations[index];
            if (!R.isNil(entity)) {
                if (entity.newEntity) {
                    result[entity.index] = entity;
                }
                else {
                    result[cursor_1++] = entity;
                }
            }
        });
    }
    return result;
};
/**
 * @stable [15.05.2019]
 * @param {MultiFieldEntityT} entity
 * @param {boolean} sort
 * @returns {TItem[]}
 */
exports.toActualMultiItemEntities = function (entity, sort) {
    if (sort === void 0) { sort = true; }
    if (R.isNil(entity)) {
        return definitions_interface_1.UNDEF;
    }
    var multiEntity = entity;
    if (!exports.isNotMultiEntity(entity)) {
        var originalSourceItems_1 = multiEntity.source || [];
        var cachedOriginalSourceItems_1 = new Map();
        originalSourceItems_1.forEach(function (itm) { return cachedOriginalSourceItems_1.set(itm.id, itm); });
        // Pass a map to optimize
        var editedEntities = exports.toActualMultiItemEditedEntities(entity, cachedOriginalSourceItems_1);
        // Remove the touched entities
        multiEntity.remove.forEach(function (itm) { return cachedOriginalSourceItems_1.delete(itm.id); });
        multiEntity.edit.forEach(function (itm) { return cachedOriginalSourceItems_1.delete(itm.id); });
        // The final snapshot
        var entities = Array.from(cachedOriginalSourceItems_1.values())
            .concat(multiEntity.add)
            .concat(editedEntities);
        if (!sort) {
            return entities;
        }
        var cachedIndexes_1 = new Map();
        entities.forEach(function (entity0) { return cachedIndexes_1.set(entity0.id, originalSourceItems_1.findIndex(function (i) { return i.id === entity0.id; })); });
        // Finally, need to sort by original entity position in an original snapshot because of added and removed entities
        return R.sort(function (item1, item2) {
            var id1 = item1.id;
            var id2 = item2.id;
            if (id1 === id2) {
                return 0;
            }
            var index1 = cachedIndexes_1.get(id1);
            var index2 = cachedIndexes_1.get(id2);
            if (index1 === -1 && index2 === -1) {
                return 0;
            }
            return index1 > index2 ? 1 : -1;
        }, entities);
    }
    return entity;
};
/**
 * @stable [22.11.2018]
 * @param {MultiFieldEntityT} entity
 * @returns {TItem[]}
 */
exports.toActualMultiItemDeletedEntities = function (entity) {
    if (R.isNil(entity)) {
        return definitions_interface_1.UNDEF;
    }
    var multiEntity = entity;
    if (!exports.isNotMultiEntity(entity)) {
        return multiEntity.remove;
    }
    return [];
};
/**
 * @stable [18.08.2018]
 * @param {MultiFieldEntityT} entity
 * @param {Map<EntityIdT, TItem extends IEntity>} mappedSourcedItems
 * @returns {TItem[]}
 */
exports.toActualMultiItemEditedEntities = function (entity, mappedSourcedItems) {
    if (R.isNil(entity)) {
        return definitions_interface_1.UNDEF;
    }
    var multiEntity = entity;
    if (!exports.isNotMultiEntity(entity)) {
        var isMappedSourcedItemsPassed = !R.isNil(mappedSourcedItems);
        var resultItems_1 = new Map();
        var cachedSourceItems_1 = mappedSourcedItems || new Map();
        if (!isMappedSourcedItemsPassed) {
            multiEntity.source.forEach(function (originalItem) { return cachedSourceItems_1.set(originalItem.id, originalItem); });
        }
        multiEntity.edit.forEach(function (editedItem) {
            var editedItemId = editedItem.id;
            var cachedResultItem = resultItems_1.get(editedItemId);
            // Collect the changes
            var editedItem0 = cachedResultItem || util_1.shallowClone(cachedSourceItems_1.get(editedItemId));
            editedItem0[editedItem.name] = editedItem.value;
            if (R.isNil(cachedResultItem)) {
                resultItems_1.set(editedItemId, editedItem0);
            }
        });
        return Array.from(resultItems_1.values());
    }
    return [];
};
/**
 * @stable [26.12.2018]
 * @param {MultiFieldEntityT} entity
 * @returns {TItem[]}
 */
exports.toActualMultiItemAddedEntities = function (entity) {
    if (R.isNil(entity)) {
        return definitions_interface_1.UNDEF;
    }
    var multiEntity = entity;
    return exports.isNotMultiEntity(entity) ? [] : multiEntity.add;
};
/**
 * @stable [29.07.2019]
 * @param {MultiFieldSingleValueT} multiFieldEntity
 * @returns {EntityIdT}
 */
exports.toLastAddedMultiItemEntityId = function (multiFieldEntity) {
    return util_1.ifNotNilThanValue(multiFieldEntity, function () { return (util_1.isPrimitive(multiFieldEntity)
        ? multiFieldEntity
        : (util_1.ifNotNilThanValue(multiFieldEntity.add, function (add) { return util_1.ifNotNilThanValue(R.last(add), function (lastValue) { return lastValue.id; }); }))); });
};
/**
 * @stable [26.06.2018]
 * @param {MultiFieldEntityT} multiFieldEntity
 * @returns {EntityIdT[]}
 */
exports.fromMultiFieldEntityToEntitiesIds = function (multiFieldEntity) {
    return fromMultiFieldEntityToEntities(multiFieldEntity, function (entity) { return entity.id; });
};
/**
 * @stable [26.06.2018]
 * @param {MultiFieldEntityT} multiFieldEntity
 * @param {(entity: TItem, index: number) => TResult} mapper
 * @returns {TResult[]}
 */
function fromMultiFieldEntityToEntities(multiFieldEntity, mapper) {
    var result = exports.toActualMultiItemEntities(multiFieldEntity);
    return util_1.orUndef(!R.isNil(result), function () { return result.map(mapper); });
}
exports.fromMultiFieldEntityToEntities = fromMultiFieldEntityToEntities;
/**
 * @stable [04.07.2018]
 * @param {MultiFieldEntityT} multiFieldEntity
 * @returns {EntityIdT[]}
 */
exports.fromMultiFieldEntityToEditedEntitiesIds = function (multiFieldEntity) {
    return fromMultiFieldEntityToEditedEntities(multiFieldEntity, function (entity) { return entity.id; });
};
/**
 * @stable [18.08.2018]
 * @param {MultiFieldEntityT} multiFieldEntity
 * @param {(entity: TItem, index: number) => TResult} mapper
 * @returns {TResult[]}
 */
function fromMultiFieldEntityToEditedEntities(multiFieldEntity, mapper) {
    var result = exports.toActualMultiItemEditedEntities(multiFieldEntity);
    return util_1.orUndef(!R.isNil(result), function () { return result.map(mapper); });
}
exports.fromMultiFieldEntityToEditedEntities = fromMultiFieldEntityToEditedEntities;
/**
 * @stable [22.11.2018]
 * @param {MultiFieldEntityT} multiFieldEntity
 * @param {(entity: TItem, index: number) => TResult} mapper
 * @returns {TResult[]}
 */
function fromMultiFieldEntityToDeletedEntities(multiFieldEntity, mapper) {
    var result = exports.toActualMultiItemDeletedEntities(multiFieldEntity);
    return util_1.orUndef(!R.isNil(result), function () { return result.map(mapper); });
}
exports.fromMultiFieldEntityToDeletedEntities = fromMultiFieldEntityToDeletedEntities;
/**
 * @stable [24.06.2018]
 * @param {MultiFieldEntityT | EntityIdT} value
 * @returns {number}
 */
exports.toActualMultiItemEntitiesLength = function (value) {
    return util_1.isDef(value)
        ? (exports.isNotMultiEntity(value)
            ? exports.normalizeEntities(value)
            : exports.toActualMultiItemEntities(value, false)).length
        : 0;
};
/**
 * @stable [03.07.2018]
 * @param {NotMultiFieldEntityT} value
 * @returns {IEntity[]}
 */
exports.normalizeEntities = function (value) {
    return util_1.isPrimitive(value) ? [{ id: value }] : value;
};
/**
 * @stable [24.06.2018]
 * @param {MultiFieldEntityT | EntityIdT} value
 * @returns {boolean}
 */
exports.isNotMultiEntity = function (value) {
    return Array.isArray(value) || util_1.isPrimitive(value);
};
/**
 * @stable [23.06.2018]
 * @param {MultiFieldEntityT} value
 * @param {(value: IMultiEntity) => IMultiItemEntity[]} converter
 * @param {IEntity[]} defaultValue
 * @returns {IMultiItemEntity[] | IEntity[]}
 */
exports.extractMultiItemEntities = function (value, converter, defaultValue) {
    return exports.isNotMultiEntity(value)
        ? (util_1.isDef(defaultValue)
            ? defaultValue
            : exports.normalizeEntities(value))
        : (R.isNil(value) ? [] : converter(value));
};
/**
 * @stable [23.06.2018]
 * @param {MultiFieldEntityT} value
 * @param {IMultiItemEntity[]} defaultValue
 * @returns {IMultiItemEntity[]}
 */
exports.extractMultiEditItemEntities = function (value, defaultValue) {
    if (defaultValue === void 0) { defaultValue = []; }
    return exports.extractMultiItemEntities(value, function (currentValue) { return currentValue.edit; }, defaultValue);
};
/**
 * @stable [23.06.2018]
 * @param {MultiFieldEntityT} value
 * @param {IMultiItemEntity[]} defaultValue
 * @returns {IMultiItemEntity[]}
 */
exports.extractMultiRemoveItemEntities = function (value, defaultValue) {
    if (defaultValue === void 0) { defaultValue = []; }
    return exports.extractMultiItemEntities(value, function (currentValue) { return currentValue.remove; }, defaultValue);
};
/**
 * @stable [02.07.2018]
 * @param {MultiFieldEntityT} value
 * @param {IMultiItemEntity[]} defaultValue
 * @returns {IEntity[]}
 */
exports.extractMultiAddItemEntities = function (value, defaultValue) {
    if (defaultValue === void 0) { defaultValue = []; }
    return exports.extractMultiItemEntities(value, function (currentValue) { return currentValue.add; }, defaultValue);
};
/**
 * @stable [23.06.2018]
 * @param {MultiFieldEntityT} value
 * @param {IEntity[]} defaultValue
 * @returns {IEntity[]}
 */
exports.extractMultiSourceItemEntities = function (value, defaultValue) {
    return exports.extractMultiItemEntities(value, function (currentValue) { return currentValue.source; }, defaultValue);
};
/**
 * @stable [02.07.2018]
 * @param {IEntity | IMultiItemEntity} entity
 * @returns {IEntity}
 */
exports.fromMultiItemEntityToEntity = function (entity) {
    var _a;
    var entityAsMultiItemEntity = entity;
    if (util_1.isDef(entityAsMultiItemEntity.rawData)) {
        return __assign({}, entityAsMultiItemEntity.rawData, (_a = {}, _a[entityAsMultiItemEntity.name] = entityAsMultiItemEntity.value, _a));
    }
    return entity;
};
/**
 * @stable [02.07.2018]
 * @param {string} name
 * @param {AnyT} value
 * @param {IEntity} rawData
 * @param {boolean} newEntity
 * @returns {IMultiItemEntity}
 */
exports.buildMultiEntity = function (name, value, rawData, newEntity) {
    return util_1.defValuesFilter({ id: rawData.id, value: value, name: name, rawData: rawData, newEntity: newEntity });
};
/**
 * @stable [23.06.2018]
 * @param {string} fieldName
 * @param {MultiFieldEntityT} multiFieldValue
 * @param {(itm: IMultiItemEntity) => boolean} predicate
 * @param {(itm: IMultiItemEntity) => AnyT} nextFieldValueFn
 * @returns {IMultiItemEntity}
 */
exports.buildMultiEditItemEntityPayload = function (fieldName, multiFieldValue, predicate, nextFieldValueFn) {
    var sourceMultiItemEntities = exports.extractMultiSourceItemEntities(multiFieldValue);
    var editedMultiItemEntities = exports.extractMultiEditItemEntities(multiFieldValue);
    var editedMultiItemEntity = editedMultiItemEntities.find(predicate);
    var sourceMultiItemEntity = sourceMultiItemEntities.find(predicate);
    return exports.buildMultiEntity(fieldName, nextFieldValueFn(editedMultiItemEntity, sourceMultiItemEntity), sourceMultiItemEntity);
};
/**
 * @stable [04.07.2018]
 * @param {IMultiItemEntity[]} multiItemEntities
 * @returns {IKeyValue}
 */
exports.fromMultiItemEntitiesToFieldsChanges = function (multiItemEntities) {
    return R.reduceBy(function (acc, entity) {
        var _a;
        return (__assign({}, acc, (_a = {}, _a[entity.name] = entity.value, _a)));
    }, {}, function (entity) { return String(entity.id); }, multiItemEntities);
};
/**
 * @stable [11.08.2018]
 * @param {IMultiItemEntity} first
 * @param {IMultiItemEntity} second
 * @returns {boolean}
 */
exports.isSameEntityByIdAndChangedFieldName = function (first, second) { return (first.id === second.id
    && first.name === second.name); };
/**
 * @stable [11.08.2018]
 * @param {IMultiItemEntity} item
 * @param {IEntity[]} addValue
 * @param {IMultiItemEntity[]} removeValue
 * @param {IMultiItemEntity[]} editValue
 * @param {IEntity[]} originalValue
 * @returns {IMultiFieldChangesEntity}
 */
exports.toMultiFieldChangesEntityOnEdit = function (item, addValue, removeValue, editValue, originalValue) {
    var removeArray = removeValue;
    var originalEditedItem = originalValue.find(function (originalItem0) { return originalItem0.id === item.id; });
    var editedNewItem = addValue.find(function (editedNewItem0) { return editedNewItem0.id === item.id; });
    var isEditedNewItem = !R.isNil(editedNewItem);
    var isOriginalEditedItem = !R.isNil(originalEditedItem);
    var editArray = isEditedNewItem
        ? editValue // If a user is editing a new record then returning an input value
        : ( // Otherwise, we should replace an old item with an input item
        editValue
            .filter(function (editedItem) { return !exports.isSameEntityByIdAndChangedFieldName(item, editedItem); })
            .concat(item)
            .filter(function (editedItem) {
            // Need to destroy the dirty changes, if an edited entity attribute is equal an original entity attribute
            return !isOriginalEditedItem
                || !(exports.isSameEntityByIdAndChangedFieldName(item, editedItem) && originalEditedItem[item.name] === item.value);
        }));
    var addArray = isEditedNewItem
        ? (item.value === definitions_interface_1.UNDEF // Need to destroy the added entities
            ? R.filter(function (itm) { return itm.id !== item.id; }, addValue)
            : R.map(function (newItem) { return newItem.id === item.id ? exports.fromMultiItemEntityToEntity(item) : newItem; }, addValue))
        : addValue;
    return { addArray: addArray, removeArray: removeArray, editArray: editArray };
};
/**
 * @stable [18.08.2018]
 * @param {IMultiItemEntity} item
 * @param {IEntity[]} addValue
 * @param {IMultiItemEntity[]} removeValue
 * @param {IMultiItemEntity[]} editValue
 * @param {IEntity[]} originalValue
 * @returns {IMultiFieldChangesEntity}
 */
exports.toMultiFieldChangesEntityOnDelete = function (item, addValue, removeValue, editValue, originalValue) {
    var deletedItemId = item.id;
    var addValueLenBeforeFiltering = addValue.length;
    var editArray = editValue.filter((function (addItem) { return addItem.id !== deletedItemId; }));
    var addArray = addValue.filter((function (addItem) { return addItem.id !== deletedItemId; }));
    var removeArray = removeValue;
    if (addArray.length === addValueLenBeforeFiltering) {
        // The local added items do not contain deleted item
        var deletedEntity = originalValue.find(function (entity) { return entity.id === deletedItemId; });
        removeArray = (R.isNil(deletedEntity) ? [] : [deletedEntity]).concat(removeValue);
    }
    else {
        // We just now have removed deleted item from addValue array
    }
    return { addArray: addArray, removeArray: removeArray, editArray: editArray };
};
/**
 * @stable [07.03.2019]
 * @param {MultiFieldEntityT<TEntity extends IEntity>} multiEntity
 * @param {MultiFieldEntityT<TEntity extends IEntity>} updatedSource
 * @param {(entity: TEntity) => boolean} addFilter
 * @param {(entity: TEntity) => boolean} editFilter
 * @param {(entity: TEntity) => boolean} removeFilter
 * @returns {IMultiEntity}
 */
exports.toFilteredMultiEntity = function (multiEntity, updatedSource, addFilter, editFilter, removeFilter) {
    if (R.isNil(multiEntity) || exports.isNotMultiEntity(multiEntity)) {
        return definitions_interface_1.UNDEF;
    }
    var multiEntity0 = multiEntity;
    var add = util_1.isFn(addFilter) ? multiEntity0.add.filter(addFilter) : multiEntity0.add;
    var edit = util_1.isFn(editFilter) ? multiEntity0.edit.filter(editFilter) : multiEntity0.edit;
    var remove = util_1.isFn(removeFilter) ? multiEntity0.remove.filter(removeFilter) : multiEntity0.remove;
    if (add.length === 0 && edit.length === 0 && remove.length === 0) {
        return definitions_interface_1.UNDEF;
    }
    return util_1.toType({
        add: add,
        edit: edit,
        remove: remove,
        source: updatedSource,
    });
};
//# sourceMappingURL=multifield.support.js.map